#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Forecast data timeseries.

This module is an interface to the Salient `forecast_timeseries` API, which returns
probabilistic weather forecasts in subseasonal-to-seasonal timescales.

Command line usage example:

```
cd ~/salientsdk
# this will get a single variable in a single file:
python -m salientsdk forecast_timeseries -lat 42 -lon -73 --timescale seasonal -u username -p password
# this will get multiple variables in separate files:
python -m salientsdk forecast_timeseries -lat 42 -lon -73 -var temp,precip --timescale seasonal
```

"""

from datetime import datetime
from typing import Literal

import pandas as pd
import requests
import xarray as xr

from .constants import Format, Weights, _build_urls, _expand_comma, _validate_enum
from .location import Location
from .login_api import download_queries

Model = Literal[
    "gem",
    "clim",
    "gfs",
    "ecmwf",
    "blend",
    "salient_clim",
    "norm_5yr",
    "norm_10yr",
    "noaa_gfs",
    "noaa_gefs",
    "noaa_gefs_calib",
    "ecmwf_ens",
    "ecmwf_ens_calib",
    "ecmwf_seas5",
    "ecmwf_seas5_calib",
    "ai",
]

Variable = Literal[
    "cc",
    "cdd",
    "hdd",
    "heat_index",
    "hgt500",
    "mslp",
    "precip",
    "rh",
    "temp",
    "tmax",
    "tmin",
    "tsi",
    "wgst",
    "wind_chill",
    "wspd",
    "wspd100",
]

Field = Literal["anom", "vals", "anom_ens", "vals_ens"]

_EXCLUDE_ARGS = ["force", "session", "verify", "verbose", "destination", "strict", "loc", "kwargs"]


def forecast_timeseries(
    # API inputs -------
    loc: Location,
    date: str | list[str] = "-default",
    debias: bool = False,
    field: Field | None = "anom",
    format: Format = "nc",
    model: Model | list[Model] = "blend",
    reference_clim: str = "salient",
    timescale="all",
    variable: Variable | list[Variable] | None = "temp",
    weights: Weights = None,
    custom_quantity: str | list[str] | None = None,
    # non-API arguments ---
    destination: str = "-default",
    force: bool = False,
    session: requests.Session | None = None,
    apikey: str | None = None,
    verify: bool | None = None,
    verbose: bool = False,
    strict: bool = True,
    **kwargs,
) -> str | pd.DataFrame:
    """Get time series of S2S meteorological forecasts.

    This function is a convenience wrapper to the Salient
    [API](https://api.salientpredictions.com/v2/documentation/api/#/Forecasts/forecast_timeseries).

    Args:
        loc (Location): The location to query
        date (str | list[str]): The date the forecast was generated.
            Defaults to find the most recent forecast.
            Can be a specific date in the format `YYYY-MM-DD`.
            May be a vector of dates or a comma separated list of dates, in which case the function will return a `DataFrame`.
        debias (bool): If True, debias the gridded forecast to local observations.
            [detail](https://salientpredictions.notion.site/Debiasing-2888d5759eef4fe89a5ba3e40cd72c8f)
        field (str): The field to query, defaults to `anom` which is an anomaly value from climatology.
            Also available: `vals`, which will return absolute values without regard to climatology.
        format (str): The file format of the response.
            Defaults to `nc` which returns a multivariate NetCDF file.
            Also available: `csv` which returns a CSV file.
        model (str | list[str]): The model to query.
            Defaults to `blend`, which is the Salient blended forecast.
            May be a list of strings or comma-separated string, which downloads multiple models.
        reference_clim (str):  Reference climatology for calculating anomalies.
            Ignored when `field=vals` since there are no anomalies to calculate.
            Defaults to `salient`, which is Salient's proprietary climatology.
        timescale (str): Forecast look-ahead.
            - 'daily' is valid for models 'gem' and 'noaa_gefs'
            - `sub-seasonal` is 1-5 weeks.  Will return a coordinate `forecast_date_weekly` and
                a data variable `anom_weekly` or `vals_weekly`.
            - `seasonal` is 1-3 months.  Will return a coordinate `forecast_date_monthly` and a
                data variable `anom_monthly` or `vals_monthly`.
            - `long-range` is 1-4 quarters.  Will return a coordinate `forecast_date_quarterly` and a
                data variable `anom_quarterly` or `vals_quarterly`.
            - `all` (default) will include `sub-seasonal`, `seasonal`, and `long-range` timescales
        variable (str | list[str] | None): The variable(s) to query, defaults to `temp`
            To request multiple variables, separate them with a comma `temp,precip` or use a `list`.
            This will download one file per variable
            See the
            [Data Fields](https://salientpredictions.notion.site/Variables-d88463032846402e80c9c0972412fe60)
            documentation for a full list of available historical and forecast variables.
        custom_quantity (str | list[str] | None): Name of a previously-uploaded custom quantity definition.
            Defaults to `None`.  If specified, ignores `variable`.
        weights (Weights): Aggregation mechanism if using a `shapefile`.
            Default `None` performs no aggregation.
        destination (str): The destination directory for downloaded files.
            `-default` uses `get_file_destination()`
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided.
        verify (bool): If True (default), verify the SSL certificate.
            Defaults to use the value returned by `get_verify_ssl()`
        verbose (bool): If True (default False) print status messages
        strict (bool): If True (default) error if query is invalid.  If False,
            return `NA` for the file name and continue processing.
        **kwargs: Additional arguments to pass to the API

    Keyword Arguments:
        units (str): `SI` or `US`

    Returns:
        str | pd.DataFrame: the file name(s) of the downloaded data.
            File names are a hash of the query parameters.
            When `force=False` and the file already exists, the function will return the file name
            almost instantaneously without querying the API.
            If multiple variables, dates, or models are requested,
            returns a `DataFrame` with column `file_name` additional columns for vectorized queries.
    """
    format = _validate_enum(format, Format, name="format")
    weights = _validate_enum(weights, Weights, name="weights")
    date = _expand_comma(date, name="date", default=datetime.today().strftime("%Y-%m-%d"))
    model = _expand_comma(model, Model, default="blend")
    custom_quantity = _expand_comma(custom_quantity, name="custom_quantity", default=None)
    if custom_quantity is None:
        variable = _expand_comma(variable, Variable, "variable", default="temp")
        assert field in ["anom", "vals", "anom_ens", "vals_ens"], f"Invalid field {field}"
    else:  # Ignore these args if custom_quantity is specified
        variable = None
        field = None

    args = {k: v for k, v in {**locals(), **kwargs}.items() if k not in _EXCLUDE_ARGS}

    queries = _build_urls(
        endpoint="forecast_timeseries", args=loc.asdict(**args), destination=destination
    )

    # strict=False will return NA filename for failures
    queries["file_name"] = download_queries(
        query=queries["query"].values,
        file_name=queries["file_name"].values,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
        strict=strict,
    )

    if len(queries) == 1:
        return queries["file_name"].values[0]
    else:
        queries = queries.drop(columns="query")  # not needed anymore
        return queries


def stack_forecast(
    ds: xr.Dataset | str | list[str] | pd.DataFrame,
    compute: bool = True,
) -> xr.Dataset:
    """Align forecast_timeseries dataset(s) along `forecast_date` as a dimension.

    Args:
        ds: Input data, which can be:
            - `xr.Dataset`: A single forecast dataset with a `forecast_date` coordinate
            - `str`: Path to a single forecast dataset file
            - `list[str]`: List of paths to multiple forecast dataset files
            - `pd.DataFrame`: DataFrame with `file_name` column containing paths
        compute: If `False`, lazy-load datasets and delay computation (default `True`)

    Returns:
        xarray Dataset with `forecast_date` as a dimension and a `time` coordinate
        that depends on `forecast_date` and `lead`.

    See Also:
        `stack_history`: Similar functionality for historical data alignment.
    """
    FCST_DIM = "forecast_date"
    LEAD_DIM = "lead"
    if isinstance(ds, xr.Dataset) or isinstance(ds, xr.DataArray):
        assert FCST_DIM in ds.coords  # Ignore forecast_date_[weekly|monthly|quarterly]
        assert LEAD_DIM in ds.coords  # Ignore lead_[weekly|monthly|quarterly]
        forecast_date = pd.Timestamp(ds[FCST_DIM].values)
        leads = ds[LEAD_DIM].values
        offset = -pd.Timedelta("1D")  # In API v2, "lead" is 1-indexed. May change in the future.
        ds = ds.drop_dims("nbnds") if "nbnds" in ds.dims else ds
        time_values = forecast_date + leads + offset
        return (
            ds.drop_vars(FCST_DIM)
            .assign_coords({"time": ("lead", time_values)})
            .expand_dims(dim={FCST_DIM: [forecast_date]})
        )
    elif isinstance(ds, str):
        args = {"decode_timedelta": True}
        opened = xr.load_dataset(ds, **args) if compute else xr.open_dataset(ds, **args)
        return stack_forecast(opened, compute=compute)
    elif isinstance(ds, pd.DataFrame):
        return stack_forecast(ds.file_name, compute=compute)
    elif isinstance(ds, list) or isinstance(ds, pd.Series):
        ds = pd.Series(ds).dropna().tolist()
        stacked = xr.open_mfdataset(
            ds,
            preprocess=stack_forecast,
            concat_dim=FCST_DIM,
            combine="nested",
            decode_timedelta=True,
        )
        if compute:
            stacked = stacked.compute()
        return stacked
    else:
        raise TypeError(f"Unsupported input type: {type(ds)}")



# Copyright Salient Predictions 2025

"""Interface to Salient forecast zarr stores.

Used in conjunction with the Salient SDK

```python
import salientsdk as sk
loc_region = sk.Location(region="north-america")
ds= sk.ForecastZarr(
    location=loc_region,
    variable="temp",
    field=["anom", "vals"],
    model="blend",
    timescale="sub-seasonal",
).open_dataset()
```
"""

import os
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable

import dask
import pandas as pd
import s3fs
import xarray as xr

from .gem_utils import get_gem_dataset
from .location import Location
from .login_api import _get_secret

MEMORY_WARN_GB = 8
MEMORY_MAX_GB = 64


class ExcessiveMemoryRequestError(Exception):
    """Exception raised when a memory request exceeds the allowed threshold for loading data."""

    pass


class ForecastZarr:
    """Factory class for accessing Salient forecast zarr stores."""

    def __new__(
        _cls,
        location: Location,
        field: str | list[str] = "anom",
        model: str | list[str] = "gem",
        timescale: str = "daily",
        variable: str = "temp",
        start: str | pd.Timestamp | None = None,
        end: str | pd.Timestamp | None = None,
        key_id: str | None = None,
        key_secret: str | None = None,
        direct_url: str | None = None,
    ):
        """Create and return an appropriate zarr accessor instance.

        Factory method that determines which zarr accessor class to instantiate
        based on the provided model parameter. Returns ZarrGEM for "gem" and
        "baseline" models, otherwise returns ZarrV9.

        Args:
            location: The Salient SDK location object.
            field: The forecast field or fields to retrieve.
            model: Which model to retrieve.
            timescale: The forecast timescale.
            variable: The forecast variable.
            start: The starting forecast date.
            end: The ending forecast date.
            key_id: The Key ID used to access the zarr stores.
            key_secret: The Key Secret used to access the zarr stores.
            direct_url: The URL for the Salient zarr store.

        Returns:
            An instance of ZarrGEM or ZarrV9, depending on the model.
        """
        start = pd.Timestamp(start) if start else None
        end = pd.Timestamp(end) if end else None
        if model == "gem":
            return ZarrGEM(
                location=location,
                field=field,
                model=model,
                timescale=timescale,
                variable=variable,
                start=start,
                end=end,
                key_id=key_id,
                key_secret=key_secret,
                direct_url=direct_url,
            )
        elif model in ["gemv1", "baseline"]:
            return ZarrGEMv1(
                location=location,
                field=field,
                model=model,
                timescale=timescale,
                variable=variable,
                start=start,
                end=end,
                key_id=key_id,
                key_secret=key_secret,
                direct_url=direct_url,
            )
        else:
            return ZarrV9(
                location=location,
                field=field,
                model=model,
                timescale=timescale,
                variable=variable,
                start=start,
                end=end,
                key_id=key_id,
                key_secret=key_secret,
                direct_url=direct_url,
            )


class ZarrBase(ABC):
    """Abstract base class for accessing Salient forecast zarr stores.

    Provides common functionality and defines the interface that concrete zarr
    accessor classes must implement. Handles parameter validation, credential
    management, location subsetting, and memory checking operations.
    """

    @property
    @abstractmethod
    def FORECAST_VARIABLES(self) -> list[str]:
        """Available forecast variables for this zarr store type."""
        pass

    @property
    @abstractmethod
    def MODELS(self) -> list[str]:
        """Available forecast models for this zarr store type."""
        pass

    @property
    @abstractmethod
    def REGIONS(self) -> list[str]:
        """Available geographic regions for this zarr store type."""
        pass

    @property
    @abstractmethod
    def FIELDS(self) -> list[str]:
        """Available forecast fields for this zarr store type."""
        pass

    @property
    @abstractmethod
    def TIMESCALES(self) -> list[str]:
        """Available forecast timescales for this zarr store type."""
        pass

    def __init__(
        self,
        location: Location,
        field: str | list[str] = "anom",
        model: str | list[str] = "blend",
        timescale: str = "sub-seasonal",
        variable: str = "temp",
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        key_id: str | None = None,
        key_secret: str | None = None,
        direct_url: str | None = None,
    ):
        """Initialize the ForecastZarr object.

        Args:
            location: The Salient SDK location object
            field: The forecast field or fields to retrieve.
            model: Which model to retrieve.
            timescale: The forecast timescale, one of `sub-seasonal`, `seasonal`, or `long-range`.
            variable: The forecast variable.
            start: The starting forecast date.
            end: The ending forecast date.
            key_id: The Key ID used to access the zarr stores. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_ID`.
            key_secret: The Key Secret used to access the zarr stores. If not provided will attempt
                to read from the environment variable `SALIENT_DIRECT_SECRET`.
            direct_url: The URL for the Salient zarr store. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_URL`.
        """
        self._validate_credentials(key_id, key_secret, direct_url)
        self.location = location
        self.field = field
        self.model = model
        self.variable = variable
        self.timescale = timescale
        self.start = start
        self.end = end

        self._validate_forecast_params()

    def _validate_forecast_params(self):
        if isinstance(self.field, str):
            assert (
                self.field in self.FIELDS
            ), f"Invalid field: {self.field}. Available fields: {self.FIELDS}"
        else:
            assert all(
                [f in self.FIELDS for f in self.field]
            ), f"Invalid field: {self.field}. Available fields: {self.FIELDS}"

        assert (
            self.model in self.MODELS
        ), f"Invalid model: {self.model}. Available models: {self.MODELS}"

        variables = self.variable if isinstance(self.variable, list) else [self.variable]
        assert all(
            [v in self.FORECAST_VARIABLES for v in variables]
        ), f"Invalid variable: {self.variable}. Available variables: {self.FORECAST_VARIABLES}"

        assert (
            self.timescale in self.TIMESCALES
        ), f"Invalid timescale: {self.timescale}. Available timescales: {self.TIMESCALES}"

        if self.start:
            self.end = self.start if self.end is None else pd.Timestamp(self.end)
            assert pd.Timestamp(self.start) <= pd.Timestamp(
                self.end
            ), f"The end date ({self.end}) must be at or after the start date ({self.start})."

        if self.location.region:
            assert (
                self.location.region in self.REGIONS
            ), f"Invalid region: {self.location.region}. Available regions: {self.REGIONS}"
            self.region = self.location.region
        else:
            self.region = "north-america"

    def _validate_credentials(
        self, key_id: str | None, key_secret: str | None, direct_url: str | None
    ):
        """Validate the credentials for direct zarr store access.

        Will read from environment variables if no arguments are passed.
        If environment variables are not found, will attempt to get secrets from Google Cloud Secret Manager.
        """
        if not key_id:
            key_id = os.getenv("SALIENT_DIRECT_ID")
            if key_id is None:
                try:
                    key_id = _get_secret("CLOUDFLARE_R2_TEST_ID")
                except Exception:
                    pass
            assert (
                key_id is not None
            ), "Must provide `key_id` or set `SALIENT_DIRECT_ID` in your environment."
        if not key_secret:
            key_secret = os.getenv("SALIENT_DIRECT_SECRET")
            if key_secret is None:
                try:
                    key_secret = _get_secret("CLOUDFLARE_R2_TEST_SECRET")
                except Exception:
                    pass
            assert (
                key_secret is not None
            ), "Must provide `key_secret` or set `SALIENT_DIRECT_SECRET` in your environment."
        if not direct_url:
            direct_url = os.getenv("SALIENT_DIRECT_URL")
            if direct_url is None:
                try:
                    direct_url = _get_secret("CLOUDFLARE_R2_TEST_URL")
                except Exception:
                    pass
            assert (
                direct_url is not None
            ), "Must provide `direct_url` or set `SALIENT_DIRECT_URL` in your environment."

        self._key_id = key_id
        self._key_secret = key_secret
        self._direct_url = direct_url

    @abstractmethod
    def open_dataset(self, in_memory: bool = False) -> xr.Dataset:
        """Open and return the forecast dataset.

        Args:
            in_memory: Whether to load the dataset into memory immediately.

        Returns:
            The xarray Dataset containing forecast data.
        """
        pass

    def subset_location(self, ds):
        """Subset the dataset to the passed `Location` object."""
        if self.location.location_file:
            df = self.location.load_location_file()
            lat, lon = ZarrBase.make_coords_dataarrays(df.lat, df.lon, df.name)
            return ZarrBase.interp_to_coords(ds, lat, lon)
        elif self.location.lat and self.location.lon:
            lat, lon = ZarrBase.make_coords_dataarrays(self.location.lat, self.location.lon)
            return ZarrBase.interp_to_coords(ds, lat, lon)
        elif self.location.region:
            return ds
        elif self.location.shapefile:
            raise NotImplementedError("Shapefile subsetting not yet implemented")
        else:
            raise ValueError("Location object must have either lat/lon, location_file, or region")

    @staticmethod
    def interp_to_coords(ds, lat=[], lon=[], method="linear"):
        """Interpolate the dataset to lat/lon locations."""
        # Interpolate to subgrid scale locations, either with sel or interp
        if method == "nearest":
            interpolated = ds.sel(lat=lat, lon=lon, method="nearest")
            interpolated = interpolated.assign_coords({"lon": lon, "lat": lat})
        else:
            # # xarray interp does sequential nd interpolation, so currently can't do this version without blowing up memory:
            # interpolated = ds.interp(lat=lat, lon=lon)
            # Instead, do a looping version through each location
            nonspatial_vars = None
            if isinstance(ds, xr.Dataset):
                # Concat tries to add location coord to nonspatial vars so separate those out
                nonspatial_vars = [
                    v for v in ds.data_vars if "lon" not in ds[v].dims and "lat" not in ds[v].dims
                ]
                ds_nonspatial = ds[nonspatial_vars]
                ds = ds.drop_vars(nonspatial_vars)
            interpolated = []
            for loc in lat.location.values:
                interpolated.append(
                    dask.delayed(ds.interp)(
                        lat=lat.sel(location=[loc]), lon=lon.sel(location=[loc]), method=method
                    )
                )
            interpolated = xr.concat(dask.compute(*interpolated), dim="location")
            if nonspatial_vars:
                interpolated = xr.merge([interpolated, ds_nonspatial])

        return interpolated

    @staticmethod
    def make_coords_dataarrays(lat, lon, names=None):
        """Make a coordinate data array from the provided location lat/lons."""
        if not isinstance(lat, Iterable):
            lat = [lat]
        if not isinstance(lon, Iterable):
            lon = [lon]
        lat = xr.DataArray(lat, dims="location", name="lat", attrs={"long_name": "Latitude"})
        lon = xr.DataArray(lon, dims="location", name="lon", attrs={"long_name": "Longitude"})
        if names is not None:
            lat = lat.assign_coords({"location": names})
            lon = lon.assign_coords({"location": names})
        return lat, lon

    def _check_memory(self, ds: xr.Dataset):
        num_gb = ds.nbytes / 1e9
        if num_gb > MEMORY_MAX_GB:
            raise ExcessiveMemoryRequestError(
                f"Loading {num_gb:,.1f} GB of data into memory exceeds the maximum allowed ({MEMORY_MAX_GB} GB)."
            )
        elif num_gb > MEMORY_WARN_GB:
            warnings.warn(
                f"Loading {num_gb:,.1f} GB of data into memory. Consider spatial selection to conserve memory and bandwidth."
            )


class ZarrV9(ZarrBase):
    """Access the native temporal resolution Salient forecast zarr stores.

    Used in order to get dask-backed Xarray datasets to the hindcast catalog of
    Salient forecasts in order to perform large-scale analyses without needing
    to make API calls.

    The available timescales are `sub-seasonal`, `seasonal`, and `long-range`.
    """

    _FORECAST_VARIABLES = ["temp", "precip", "wspd", "tsi", "cdd", "hdd"]
    _MODELS = ["blend", "noaa_gefs", "ecmwf_ens", "ecmwf_seas5", "truth"]
    _REGIONS = ["north-america"]
    _FIELDS = ["anom", "vals"]
    _TIMESCALES = ["sub-seasonal", "seasonal", "long-range"]

    @property
    def FORECAST_VARIABLES(self):
        """Available forecast variables for this zarr store type."""
        return self._FORECAST_VARIABLES

    @property
    def MODELS(self):
        """Available forecast models for this zarr store type."""
        return self._MODELS

    @property
    def REGIONS(self):
        """Available geographic regions for this zarr store type."""
        return self._REGIONS

    @property
    def FIELDS(self):
        """Available forecast fields for this zarr store type."""
        return self._FIELDS

    @property
    def TIMESCALES(self):
        """Available forecast timescales for this zarr store type."""
        return self._TIMESCALES

    def __init__(
        self,
        location: Location,
        field: str | list[str] = "anom",
        model: str | list[str] = "blend",
        timescale: str = "sub-seasonal",
        variable: str = "temp",
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        key_id: str | None = None,
        key_secret: str | None = None,
        direct_url: str | None = None,
    ):
        """Initialize the ForecastZarr object.

        Args:
            location: The Salient SDK location object
            field: The forecast field or fields to retrieve.
            model: Which model to retrieve.
            timescale: The forecast timescale, one of `sub-seasonal`, `seasonal`, or `long-range`.
            variable: The forecast variable.
            start: The starting forecast date.
            end: The ending forecast date.
            key_id: The Key ID used to access the zarr stores. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_ID`.
            key_secret: The Key Secret used to access the zarr stores. If not provided will attempt
                to read from the environment variable `SALIENT_DIRECT_SECRET`.
            direct_url: The URL for the Salient zarr store. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_URL`.
        """
        super().__init__(
            location=location,
            field=field,
            model=model,
            timescale=timescale,
            variable=variable,
            start=start,
            end=end,
            key_id=key_id,
            key_secret=key_secret,
            direct_url=direct_url,
        )

    def open_dataset(self, in_memory: bool = False) -> xr.Dataset:
        """Open the dataset.

        Will be returned as dask-backed xr.Dataset, unless `in_memory=True`, and then the values
        will be loaded into memory.
        """
        fs = s3fs.S3FileSystem(
            key=self._key_id,
            secret=self._key_secret,
            client_kwargs=dict(endpoint_url=self._direct_url, region_name="enam"),
            s3_additional_kwargs=dict(ACL="private"),
        )

        fields = [self.field] if isinstance(self.field, str) else self.field
        if self.model == "truth":
            fields = [f"{field}_actual" for field in fields]

        store_path = f"{self.region}-{self.variable}/{self.timescale}/{self.model}"
        # Check is false since we're pointing to a directory, not an object
        store = s3fs.S3Map(root=store_path, s3=fs, check=False, create=False)
        ds = xr.open_zarr(store=store).sel(forecast_date=slice(self.start, self.end))[fields]
        ds = self.subset_location(ds)

        if in_memory:
            self._check_memory(ds)
            ds = ds.load()

        return ds


class ZarrGEM(ZarrBase):
    """Access the native Salient forecast zarr stores for the GemAI v2 global model.

    Used in order to get dask-backed Xarray datasets to the hindcast and operational catalog of
    Salient forecasts in order to perform large-scale analyses without needing
    to make API calls.
    """

    _BASE_VARS = [
        "cc",
        "dewpoint",
        "hgt500",
        "mslp",
        "precip",
        "tmax",
        "tmin",
        "tsi",
        "wgst",
        "wspd",
        "wspd100",
    ]
    _DERIVED_VARS = ["cdd", "hdd", "heat_index", "wind_chill", "temp", "rh"]
    _FORECAST_VARIABLES = _BASE_VARS + _DERIVED_VARS

    _MODELS = ["gem"]
    _REGIONS = [
        "africa",
        "asia",
        "europe",
        "global",
        "north-america",
        "russia",
        "south-america",
        "south-pacific",
    ]
    _FIELDS = ["vals", "vals_ens"]
    _TIMESCALES = ["daily"]
    _AVAILABLE_DATES = xr.date_range("2020-01-01", "today", freq="D")

    @property
    def FORECAST_VARIABLES(self):
        """Available forecast variables for this zarr store type."""
        return self._FORECAST_VARIABLES

    @property
    def MODELS(self):
        """Available forecast models for this zarr store type."""
        return self._MODELS

    @property
    def REGIONS(self):
        """Available geographic regions for this zarr store type."""
        return self._REGIONS

    @property
    def FIELDS(self):
        """Available forecast fields for this zarr store type."""
        return self._FIELDS

    @property
    def TIMESCALES(self):
        """Available forecast timescales for this zarr store type."""
        return self._TIMESCALES

    def __init__(
        self,
        location: Location,
        field: str | list[str] = "vals",
        model: str | list[str] = "gek",
        timescale: str = "daily",
        variable: str = "tmax",
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        key_id: str | None = None,
        key_secret: str | None = None,
        direct_url: str | None = None,
    ):
        """Initialize the ForecastZarr object.

        Args:
            location: The Salient SDK location object
            field: The forecast field or fields to retrieve.
            model: Which model to retrieve.
            timescale: The forecast timescale, one of `sub-seasonal`, `seasonal`, or `long-range`.
            variable: The forecast variable.
            start: The starting forecast date.
            end: The ending forecast date.
            key_id: The Key ID used to access the zarr stores. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_ID`.
            key_secret: The Key Secret used to access the zarr stores. If not provided will attempt
                to read from the environment variable `SALIENT_DIRECT_SECRET`.
            direct_url: The URL for the Salient zarr store. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_URL`.
        """
        super().__init__(
            location=location,
            field=field,
            model=model,
            timescale=timescale,
            variable=variable,
            start=start,
            end=end,
            key_id=key_id,
            key_secret=key_secret,
            direct_url=direct_url,
        )

    def open_dataset(self, in_memory: bool = False) -> xr.Dataset:
        """Open the dataset.

        Will be returned as dask-backed xr.Dataset, unless `in_memory=True`, and then the values
        will be loaded into memory.
        """
        if not self.start:
            self.start = pd.Timestamp(self._AVAILABLE_DATES.min())
        if not self.end:
            self.end = pd.Timestamp(self._AVAILABLE_DATES.max())

        if isinstance(self.field, str):
            self.field = [self.field]
        ds = xr.merge(
            [
                get_gem_dataset(
                    variables=self.variable,
                    model=self.model,
                    field=field,
                    region=self.region,
                    start=self.start,
                    end=self.end,
                    key_id=self._key_id,
                    key_secret=self._key_secret,
                    direct_url=self._direct_url,
                )
                for field in self.field
            ]
        )
        ds = self.subset_location(ds)

        if in_memory:
            self._check_memory(ds)
            ds = ds.load()

        return ds


class ZarrGEMv1(ZarrBase):
    """Access the native Salient forecast zarr stores.

    Used in order to get dask-backed Xarray datasets to the hindcast catalog of
    Salient forecasts in order to perform large-scale analyses without needing
    to make API calls.
    """

    _BASE_VARS = [
        "cc",
        "hgt500",
        "mslp",
        "precip",
        "rh",
        "tmax",
        "tmin",
        "tsi",
        "wgst",
        "wspd",
        "wspd100",
    ]
    _DERIVED_VARS = ["cdd", "hdd", "heat_index", "wind_chill", "temp"]
    _FORECAST_VARIABLES = _BASE_VARS + _DERIVED_VARS

    _MODELS = ["gemv1", "baseline"]
    _REGIONS = ["north-america"]
    _FIELDS = ["anom", "vals", "anom_ens", "vals_ens"]
    _TIMESCALES = ["daily"]
    _AVAILABLE_DATES = xr.date_range("2020-10-16", "today", freq="D")

    @property
    def FORECAST_VARIABLES(self):
        """Available forecast variables for this zarr store type."""
        return self._FORECAST_VARIABLES

    @property
    def MODELS(self):
        """Available forecast models for this zarr store type."""
        return self._MODELS

    @property
    def REGIONS(self):
        """Available geographic regions for this zarr store type."""
        return self._REGIONS

    @property
    def FIELDS(self):
        """Available forecast fields for this zarr store type."""
        return self._FIELDS

    @property
    def TIMESCALES(self):
        """Available forecast timescales for this zarr store type."""
        return self._TIMESCALES

    def __init__(
        self,
        location: Location,
        field: str | list[str] = "anom",
        model: str | list[str] = "gemv1",
        timescale: str = "daily",
        variable: str = "tmax",
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        key_id: str | None = None,
        key_secret: str | None = None,
        direct_url: str | None = None,
    ):
        """Initialize the ForecastZarr object.

        Args:
            location: The Salient SDK location object
            field: The forecast field or fields to retrieve.
            model: Which model to retrieve.
            timescale: The forecast timescale, one of `sub-seasonal`, `seasonal`, or `long-range`.
            variable: The forecast variable.
            start: The starting forecast date.
            end: The ending forecast date.
            key_id: The Key ID used to access the zarr stores. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_ID`.
            key_secret: The Key Secret used to access the zarr stores. If not provided will attempt
                to read from the environment variable `SALIENT_DIRECT_SECRET`.
            direct_url: The URL for the Salient zarr store. If not provided will attempt to read
                from the environment variable `SALIENT_DIRECT_URL`.
        """
        super().__init__(
            location=location,
            field=field,
            model=model,
            timescale=timescale,
            variable=variable,
            start=start,
            end=end,
            key_id=key_id,
            key_secret=key_secret,
            direct_url=direct_url,
        )

    def open_dataset(self, in_memory: bool = False) -> xr.Dataset:
        """Open the dataset.

        Will be returned as dask-backed xr.Dataset, unless `in_memory=True`, and then the values
        will be loaded into memory.
        """
        if not self.start:
            self.start = pd.Timestamp(self._AVAILABLE_DATES.min())
        if not self.end:
            self.end = pd.Timestamp(self._AVAILABLE_DATES.max())

        if isinstance(self.field, str):
            self.field = [self.field]
        ds = xr.merge(
            [
                get_gem_dataset(
                    variables=self.variable,
                    model=self.model,
                    field=field,
                    region=self.region,
                    start=self.start,
                    end=self.end,
                    key_id=self._key_id,
                    key_secret=self._key_secret,
                    direct_url=self._direct_url,
                )
                for field in self.field
            ]
        )
        ds = self.subset_location(ds)

        if in_memory:
            self._check_memory(ds)
            ds = ds.load()

        return ds



#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Observed data timeseries.

This module acquires observed station meteorological data and converts it into a format
compatible with the `data_timeseries` function.
"""

from collections.abc import Iterable
from typing import Literal

import numpy as np
import pandas as pd
import requests
import xarray as xr

from .constants import _build_urls, _collapse_comma
from .location import Location
from .login_api import download_queries

_EXCLUDE_ARGS = ["force", "session", "verify", "verbose", "destination", "loc", "strict", "kwargs"]


MetVariable = Literal["precip", "temp", "tmax", "tmin", "snow", "wdir", "wspd"]


def met_stations(
    # API arguments -----
    loc: Location,
    variables: MetVariable | list[MetVariable] | None = None,
    start: str | None = "2020-01-01",
    end: str | None = None,
    max_distance: float = 10,
    # Non-API arguments --------
    destination: str = "-default",
    force: bool = False,
    session: requests.Session | None = None,
    apikey: str | None = None,
    verify: bool | None = None,
    verbose: bool = False,
    strict: bool = True,
    **kwargs,
) -> str | pd.DataFrame:
    """Search meteorological stations.

    For more detail, see the
    [api doc](https://api.salientpredictions.com/v2/documentation/api/#/Meteorological%20Stations/met_stations).

    Args:
        loc (Location): The location to query.
            If using a `shapefile` or `location_file`, may input a vector of file names which
            will trigger multiple calls to `downscale`.  This is useful because `downscale` requires
            that all points in a file be from the same continent.
        variables (MetVariable | list[MetVariable] | None): The variables that must be available,
            separated by commas or as a `list`.
            Must be one or more of precip, temp, tmax, tmin, snow, wdir, or wspd.
            Not all data is available at all stations.
            If omitted (default), does not filter stations by variable availability.
        start (str): Makes sure that data is available
            at least as long ago as this date.  If omitted, does not filter station list
            by availability date.
        end (str): Makes sure that data is available at least as recently as this date.
            If omitted (default) does not filter station list by availability date.
        max_distance (float): The maximum allowable distance in kilometers between a coordinate
            and its nearest station. If this distance is exceeded, no station will be returned
            for the coordinate. Only applies when `lat/lon` or `location_file` is specified.
        destination (str): The destination directory for downloaded files.
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        strict (bool): If True (default) error if query is invalid.  If False,
            return `NA` for the file name and continue processing.
        **kwargs: Additional arguments to pass to the API

    Returns:
        str | pd.DataFrame : If only one file was downloaded, return the name of the file.
            If multiple files were downloaded, return a table with column `file_name` and
            additional columns documenting the vectorized input arguments such as
            `location_file`.
    """
    format = "csv"
    variables = _collapse_comma(variables, valid=MetVariable)
    assert max_distance > 0
    args = {k: v for k, v in {**locals(), **kwargs}.items() if k not in _EXCLUDE_ARGS}

    queries = _build_urls(
        endpoint="met_stations", args=loc.asdict(**args), destination=destination
    )

    queries["file_name"] = download_queries(
        query=queries["query"].values,
        file_name=queries["file_name"].values,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
        strict=strict,
    )

    if len(queries) == 1:
        return queries["file_name"].values[0]
    else:
        queries = queries.drop(columns="query")
        return queries


def met_observations(
    # API arguments -----
    loc: Location,
    variables: MetVariable | list[MetVariable] = ["tmax", "tmin", "precip"],
    start: str | None = "2020-01-01",
    end: str | None = None,
    max_distance: float = 10,
    format: str = "nc",
    # Non-API arguments --------
    destination: str = "-default",
    force: bool = False,
    session: requests.Session | None = None,
    apikey: str | None = None,
    verify: bool | None = None,
    verbose: bool = False,
    strict: bool = True,
    **kwargs,
) -> str | pd.DataFrame:
    """Fetch meteorological station observations.

    For more detail, see the
    [api doc](https://api.salientpredictions.com/v2/documentation/api/#/Meteorological%20Stations/met_observations).

    Args:
        loc (Location): The location to query.
            If using a `shapefile` or `location_file`, may input a vector of file names which
            will trigger multiple calls to `downscale`.  This is useful because `downscale` requires
            that all points in a file be from the same continent.
        variables (MetVariable | list[MetVariable] | None): The variables to download,
            separated by commas or as a `list`.
            Must be one or more of precip, temp, tmax, tmin, snow, wdir, or wspd.
            Not all data is available at all stations.
        start (str): Makes sure that data is available
            at least as long ago as this date.  If omitted, does not filter station list
            by availability date.
        end (str): Makes sure that data is available at least as recently as this date.
            If omitted (default) does not filter station list by availability date.
        max_distance (float): The maximum allowable distance in kilometers between a coordinate
            and its nearest station. If this distance is exceeded, no station will be returned
            for the coordinate. Only applies when `lat/lon` or `location_file` is specified.
        format (str): `nc` or `csv`
        destination (str): The destination directory for downloaded files.
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        strict (bool): If True (default) error if query is invalid.  If False,
            return `NA` for the file name and continue processing.
        **kwargs: Additional arguments to pass to the API

    Returns:
        str | pd.DataFrame : If only one file was downloaded, return the name of the file.
            If multiple files were downloaded, return a table with column `file_name` and
            additional columns documenting the vectorized input arguments such as
            `location_file`.
    """
    variables = _collapse_comma(variables, valid=MetVariable)
    assert max_distance > 0
    args = {k: v for k, v in {**locals(), **kwargs}.items() if k not in _EXCLUDE_ARGS}

    queries = _build_urls(
        endpoint="met_observations", args=loc.asdict(**args), destination=destination
    )

    queries["file_name"] = download_queries(
        query=queries["query"].values,
        file_name=queries["file_name"].values,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
        strict=strict,
    )

    if len(queries) == 1:
        return queries["file_name"].values[0]
    else:
        queries = queries.drop(columns="query")
        return queries


def make_observed_ds(
    obs_df: pd.DataFrame | str | Iterable[pd.DataFrame | str],
    name: str | Iterable[str],
    variable: str | Iterable[str],
    time_col: str = "time",
) -> xr.Dataset:
    """Convert weather observation DataFrame(s) to xarray Dataset.

    This function converts tabular meteorological data into a format identical to
    `data_timeseries(..., frequency='daily') suitable for use with the `crps` function.

    Args:
        obs_df: Single DataFrame or a filename to a CSV that can be read as a dataframe.
            May also be an iterable vector of filenames or `DataFrame`s.
            Each DataFrame should have columns for `time` and the `variable` of interest.
            If the dataframe contains lat, lon, and elev metadata the function will
            preserve thse as coordinates. Function `get_ghcnd` will provide a compatible
            dataset, or you can provide your own.
        name: Station name(s) corresponding to the DataFrame(s). Must be a string if obs_df
            is a single DataFrame, or an iterable of strings matching the length of obs_df
            if multiple DataFrames are provided.
        variable: Name(s) of the column(s) in obs_df to extract the met data
            (e.g. 'temp', 'precip') or ['temp','precip']
        time_col: Name of the column in obs_df containing the time (default `time`)

    Returns:
        xarray Dataset containing the variable data and station metadata. Has dimensions
        'time' and 'location', with coordinates for station lat/lon/elevation.
    """

    def get_attrs(var_name):
        """Helper to collect attributes from a dataframe column."""
        attrs = {"short_name": var_name}
        if hasattr(obs_df[var_name], "attrs"):
            for attr in ["units", "long_name"]:
                if attr in obs_df[var_name].attrs:
                    attrs[attr] = obs_df[var_name].attrs[attr]
        return attrs

    if isinstance(obs_df, Iterable) and not isinstance(obs_df, pd.DataFrame):
        if name is None or isinstance(name, str):
            raise ValueError(
                "When obs_df is a list of DataFrames, name must be an iterable of strings"
            )

        assert len(obs_df) == len(
            name
        ), f"Length mismatch: got {len(obs_df)} DataFrames but {len(name)} names"

        ds = [
            make_observed_ds(obs_df=df, name=n, variable=variable, time_col=time_col)
            for df, n in zip(obs_df, name)
        ]
        return xr.concat(ds, dim="location")

    if isinstance(obs_df, str):
        obs_df = pd.read_csv(obs_df)

    name = str(name)
    attrs = {}

    if isinstance(variable, str):
        data_vars = {"vals": (("time", "location"), obs_df[variable].values[:, np.newaxis])}
        attrs = get_attrs(variable)
    else:
        data_vars = {}
        for var in variable:
            data_vars[var] = (
                ("time", "location"),
                obs_df[var].values[:, np.newaxis],
                get_attrs(var),
            )

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={
            "time": pd.to_datetime(obs_df[time_col]),
            "location": [name],
        },
        attrs=attrs,
    )

    # Preserve station geo-coordinates if they exist as a column
    for coord in ["lat", "lon", "elev"]:
        if coord in obs_df.columns:
            ds = ds.assign_coords({f"{coord}_station": ("location", [obs_df[coord].mean()])})

    return ds



"""GEM forecast data utilities and adapters.

This module provides utilities for loading, processing, and adapting GEM (Global Environmental
Multiscale) forecast data from zarr stores. It includes functions for data clipping, quantile
computation, coordinate promotion, and format adaptation for API compatibility.

Key functionality:
- Loading GEM and baseline forecast data from S3-compatible zarr stores
- Variable clipping to enforce physical constraints (e.g., non-negative precipitation)
- Converting ensemble forecasts to quantile-based representations
- Adapting data formats for backwards compatibility with existing APIs
- Computing derived meteorological quantities
"""

from typing import Literal

import pandas as pd
import s3fs
import xarray as xr

from .derived import compute_quantity
from .location import _load_regions_mask

mask = _load_regions_mask()
CLIPPERS = dict(
    precip={"min": 0},
    tsi={"min": 0},
    wspd={"min": 0},
    wspd100={"min": 0},
    wgst={"min": 0},
    cc={"min": 0, "max": 1},
)
QUANTILES = [
    0.01,
    0.025,
    0.05,
    0.1,
    0.15,
    0.2,
    0.25,
    0.3,
    0.35,
    0.4,
    0.45,
    0.5,
    0.55,
    0.6,
    0.65,
    0.7,
    0.75,
    0.8,
    0.85,
    0.9,
    0.95,
    0.975,
    0.99,
]


def _create_fs(key_id: str, key_secret: str, direct_url: str) -> s3fs.S3FileSystem:
    fs = s3fs.S3FileSystem(
        key=key_id,
        secret=key_secret,
        client_kwargs=dict(endpoint_url=direct_url, region_name="enam"),
        s3_additional_kwargs=dict(ACL="private"),
    )
    return fs


def _load_baseline(
    fs: s3fs.S3FileSystem,
    region: str,
    date: str | None = None,
    chunks: dict | None = None,
) -> xr.Dataset:
    store_path = f"gem-{region}/baseline"
    # Check is false since we're pointing to a directory, not an object
    store = s3fs.S3Map(root=store_path, s3=fs, check=False, create=False)

    ds = xr.open_zarr(store, chunks=None, decode_timedelta=True).drop_vars("historical_date")
    chunks = chunks or dict(forecast_date=1, lead=7, sample=-1, lat=60, lon=60)

    # Handle date selection separately for the one year of baseline forecasts
    if date is not None:
        ds = ds.sel(forecast_date=[pd.Timestamp(date).replace(year=2024)])
        ds = ds.drop_vars("forecast_date").assign_coords(forecast_date=[pd.Timestamp(date)])

    ds = _subselect_and_chunk(ds, None, chunks=chunks)

    return ds


def _load_gem(
    fs: s3fs.S3FileSystem,
    region: str,
    date: str | None = None,
    start: str | None = None,
    end: str | None = None,
    chunks: dict | None = None,
    model: str = "gem",
) -> xr.Dataset:
    if model == "gemv1":
        store_path = f"gem-{region}/forecast"
        store_chunks = dict(forecast_date=1, lead=7, sample=-1, lat=60, lon=60)
    elif model == "gem":
        store_path = "gemv2/forecast"
        store_chunks = dict(forecast_date=1, lead=7, sample=-1, lat=144, lon=144)
    else:
        raise ValueError(f"Model {model} not supported, must be 'gem' or 'gemv1'")

    store = s3fs.S3Map(root=store_path, s3=fs, check=False, create=False)
    ds = xr.open_zarr(store, chunks=None, decode_timedelta=True)
    chunks = chunks or store_chunks

    ds = _subselect_and_chunk(
        ds,
        date,
        start=start,
        end=end,
        chunks=chunks,
        # Don't subset regional stores from GEMv1, but do so for GEMv2
        region=region if model == "gem" else "global",
    )
    if model == "gem":

        def _get_num_leads(d: pd.Timestamp):
            operational_start = pd.Timestamp("2025-10-01")
            if d >= operational_start or d.dayofweek == 0:
                return 126
            else:
                return 46

        num_leads = xr.zeros_like(ds.forecast_date).astype(int).chunk(forecast_date=-1)
        num_leads.values = [_get_num_leads(pd.Timestamp(fd)) for fd in ds.forecast_date.values]
        ds = ds.assign_coords(num_leads=num_leads)
    return ds


def _align_baseline(ds: xr.Dataset, baseline: xr.Dataset) -> xr.Dataset:
    """Merge the baseline dataset into the main dataset for arbitrary range of forecast dates.

    The `baseline` dataset represents a climatology, but instead of having a `dayofyear` dimension,
    it has a `forecast_date` dimension so that it can easily be used seamlessly with the GEM
    dataset. However, its forecast dates are only for 2024, so to join to a range of non-2024
    `forecast_date`s in the GEM dataset, we loop through the years present in the GEM dataset
    to easily select the correct climatology for the baseline, then concatenate these datasets
    together so that the GEM and baseline datasets are aligned.

    Args:
        ds: The GEM dataset to align to. Can have any arbitrary range of `forecast_date`s
        baseline: The baseline dataset (i.e., climo) to join with the GEM dataset.

    Returns:
        A `baseline` dataset with the same `forecast_date`s as the GEM dataset to represent
            the climatology to be used for each `forecast_date`.
    """
    ds_years = set(ds.forecast_date.dt.year.values)
    datasets = []
    for year in sorted(ds_years):
        year_str = f"{year}"
        forecast_dates = ds.sel(forecast_date=slice(year_str, year_str)).forecast_date.values
        data = baseline.sel(
            forecast_date=[pd.Timestamp(date).replace(year=2024) for date in forecast_dates]
        )
        datasets.append(
            data.drop_vars("forecast_date").assign_coords(
                forecast_date=[pd.Timestamp(date) for date in forecast_dates]
            )
        )

    return xr.concat(datasets, dim="forecast_date").transpose(*list(ds.dims))


def _subselect_and_chunk(
    ds: xr.Dataset,
    date: str | None = None,
    start: str | None = None,
    end: str | None = None,
    chunks: dict | None = None,
    region: str = "global",
) -> xr.Dataset:
    if "ensemble" in ds.dims:
        ds = ds.rename(ensemble="sample")

    if date:
        ds = ds.sel(forecast_date=[date])
    elif start or end:
        ds = ds.sel(forecast_date=slice(start, end))

    if chunks is not None:
        ds = ds.chunk(chunks)

    if region != "global":
        mask = _load_regions_mask()
        region_map = {
            "north-america": 1,
            "south-america": 2,
            "europe": 3,
            "russia": 4,
            "africa": 5,
            "asia": 6,
            "south-pacific": 7,
        }
        if region not in region_map:
            raise ValueError(
                f"Region {region} not recognized, must be one of {list(region_map.keys())}"
            )
        ds = ds.where(mask.mask == region_map[region], drop=True)

    return ds


def _clip_variables(ds: xr.Dataset) -> xr.Dataset:
    """Clip variable to defined min/max values.

    This is used because we use a heavy compression scheme on the store, so it's possible
    to have very small (<1e-3) negative values for say, precipitation or tsi, and we don't
    want to return these invalid values.
    """
    clipped_vars = set(ds.data_vars).intersection(CLIPPERS.keys())
    if clipped_vars:
        for var in clipped_vars:
            ds[var] = ds[var].clip(**CLIPPERS[var])
    return ds


def _promote_coords(
    ds: xr.Dataset, coords_to_promote: list[str] = ["forecast_date", "lead"]
) -> xr.Dataset:
    """Promote coordinates to dimensions if they are not already dimensions."""
    for coord in coords_to_promote:
        if coord in ds.coords and coord not in ds.dims:
            ds = ds.expand_dims(coord)
    return ds


def _get_quantiles(
    ds: xr.Dataset,
    quantiles: float | list[float] = QUANTILES,
    interp_method: str = "linear",
    quantile_dim_name: str = "quantiles",
    skipna: bool = True,
) -> xr.Dataset:
    """Convert v10 forecast from sample space to quantile space.

    Running `ds.quantile` will drop any coordinates that are single-valued that are not
    also dimensions so we promote the `forecast_date` and `lead` coordinates if they are
    not already dimensions.
    """
    ds = _promote_coords(ds, coords_to_promote=["forecast_date", "lead"])
    return ds.quantile(
        quantiles, dim="sample", method=interp_method, keep_attrs=True, skipna=skipna
    ).rename({"quantile": quantile_dim_name})


def v10_adapter(ds: xr.Dataset, decode_timedelta: bool = True) -> xr.Dataset:
    """Adapt the GEM dataset to the old daily GEFS format for use in API."""
    dims = ["forecast_date", "lead"]
    dims = dims + ["lat", "lon"] if "lat" in ds.dims and "lon" in ds.dims else dims
    dims = dims + ["ensemble"] if "sample" in ds.dims else dims
    renamer = {"sample": "ensemble"} if "sample" in ds.dims else {}

    if decode_timedelta:
        leads = ds.lead
    else:
        leads = pd.to_timedelta([l for l in ds.lead.values], unit="D")

    ds_new = ds.assign_coords({"lead": leads}).rename(**renamer).transpose(*dims, ...)

    for v in ds_new.data_vars:
        anomaly = " anomaly" if "anom" in str(v) else ""
        ensembles = " (ensembles)" if "ens" in str(v) else ""
        ds_new[v].attrs["long_name"] += anomaly + ensembles

    if not decode_timedelta:
        ds_new = ds_new.assign_coords(lead=ds_new.lead.dt.days)
        ds_new["lead"].attrs["units"] = "days"

    return ds_new


def get_gem_dataset(
    variables: str | list[str],
    model: Literal["gem", "gemv1", "baseline"] = "gem",
    field: str | list[str] | None = ["vals_ens"],
    date: str | pd.Timestamp | None = None,
    start: str | pd.Timestamp | None = None,
    end: str | pd.Timestamp | None = None,
    region: str = "global",
    decode_timedelta: bool = True,
    chunks: dict | None = None,
    quantiles: float | list[float] = QUANTILES,
    adapt: bool = True,
    key_id: str | None = None,
    key_secret: str | None = None,
    direct_url: str | None = None,
    **kwargs,
) -> xr.Dataset:
    """Primary data adapter for GEM forecasts.

    The v10 dataset is stored in a "base" zarr store with all variables. Returns from this store
    will apend the field to the variable name, e.g., `temp_vals`, `precip_anom`, etc. to conform
    with other API integrations.

    Args:
        variables: Variable or list of variables to return.
        model: Model to return data for, either "gem" or "baseline".
        field: Field type to return, either "vals", "anom", or "vals_ens"/"anom_ens".
            If None, defaults to ["vals_ens"].
        date: Specific date to return data for. If None, returns all available dates.
        start: Start date for date range selection.
        end: End date for date range selection.
        region: The modeling region, currently only "north-america" supported.
        decode_timedelta: Whether to decode lead time as timedelta objects.
            If False, returns integer day values.
        chunks: Custom chunking configuration for the dataset.
        quantiles: Quantiles to compute for ensemble-to-quantile conversion.
        adapt: Whether to adapt dataset to legacy GEFS format for API usage.
        key_id: S3 access key ID for zarr store authentication.
        key_secret: S3 secret key for zarr store authentication.
        direct_url: Direct URL endpoint for S3-compatible zarr store.
        **kwargs: Additional arguments passed to loading functions.

    Returns:
        Processed xarray Dataset with requested variables and transformations applied.

    Raises:
        ValueError: If unsupported model is specified.
    """
    if isinstance(field, str):
        field = [field]

    if isinstance(variables, str):
        variables = [variables]

    fs = _create_fs(key_id, key_secret, direct_url)

    if model in ["gem", "gemv1"]:
        ds = _load_gem(
            fs, region, date=date, start=start, end=end, chunks=chunks, model=model, **kwargs
        )
    elif model == "baseline":
        ds = _load_baseline(fs, region, date=date, chunks=chunks, **kwargs)
    else:
        raise ValueError(f"Model {model} not supported")

    ds = _clip_variables(ds)
    ds_processed = xr.Dataset()
    for v in variables:
        ds_processed[v] = compute_quantity(ds, v)
        ds_processed[v].attrs.pop("kind", None)

    ds = ds_processed
    # Drop timedelta encoding after so we can compute derived quantities properly
    if not decode_timedelta:
        ds = ds.assign_coords(lead=ds.lead.dt.days)
        ds["lead"].attrs["units"] = "days"

    ds = ds.rename({v: f"{v}_vals_ens" for v in variables})

    if "anom" in field or "anom_ens" in field:
        if model == "gem":
            raise ValueError("Anomalies can only be computed for model='gemv1'")
        # Call recursively to get baseline forecast as the climo reference
        baseline = get_gem_dataset(
            variables,
            model="baseline",
            field="vals_ens",
            date=date,
            region=region,
            decode_timedelta=decode_timedelta,
            adapt=False,  # Don't adapt until we compute anoms
            key_id=key_id,
            key_secret=key_secret,
            direct_url=direct_url,
            **kwargs,
        )
        baseline = _align_baseline(ds, baseline)
        baseline_mean = baseline.mean(dim="sample", keep_attrs=True)
        with xr.set_options(keep_attrs=True):
            ds = ds - baseline_mean
        ds = ds.rename({f"{v}_vals_ens": f"{v}_anom_ens" for v in variables})

    if "anom" in field or "vals" in field:
        v_a = "vals" if "vals" in field else "anom"
        ds = _get_quantiles(ds, quantiles=quantiles, skipna=False)
        ds = ds.rename({f"{v}_{v_a}_ens": f"{v}_{v_a}" for v in variables})

    if adapt:
        ds = v10_adapter(ds, decode_timedelta=decode_timedelta)

    return ds



#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Detect and analyze events."""

from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr
from pandas.io.formats.style import Styler
from sklearn.linear_model import LogisticRegression

# Default payoff matrix
PAYOFF = {
    # 2x2 simple matrix payoff coefficients:
    "pp": 100,  #    Benefit of correctly predicting an event
    "fp": -10,  #    Cost of falsely predicting an event
    "tn": 5,  #      Benefit of correctly predicting no event
    "fn": -50,  #    Cost of missing an event
    # 4x2 joint matrix payoff coefficients:
    "npp": -10,
    "ppp": 100,  #   Both agree there will be an extreme.
    "npn": -20,
    "ppn": 300,  #   Forecast extreme, ref normal.  Differentiated alpha
    "nnp": -20,
    "pnp": 300,  #   Forecast normal, ref extreme.  Differentiated alpha.
    "nnn": 0,
    "pnn": 0,  #     Neither extreme, take no action.
    # For setting axis labels
    "units": "$",
}


def _safe_divide(
    numerator: xr.Dataset | xr.DataArray,
    denominator: xr.Dataset | xr.DataArray,
    long_name: str | None = None,
    units: str | None = None,
    **attrs,
) -> xr.Dataset | xr.DataArray:
    """Avoid divide-by-zero errors in division.

    Args:
        numerator: The dividend in the division operation
        denominator: The divisor in the division operation
        long_name: Optional name to add to the result's attributes
        units: Optional units to add to the result's attributes
        **attrs: Additional attributes to add to the result.
            Overridden by long_name and units, if present.

    Returns:
        The result of numerator/denominator, with 0.0 returned for any cases where
        denominator is zero to avoid division errors. Return type matches input type.
    """
    with xr.set_options(keep_attrs=True):
        # the double-where idiom here is necessary to be dask-friendly
        masked_denominator = xr.where(denominator == 0, 1.0, denominator)
        result = numerator / masked_denominator
        result = xr.where(denominator != 0, result, 0.0)

    attrs = attrs.copy()  # prevent changing by reference
    if long_name is not None:
        attrs["long_name"] = long_name
    if units is not None:
        attrs["units"] = units
    if attrs:
        result.attrs.update(attrs)

    return result


def classify_event(
    x: xr.DataArray,
    target: float | xr.DataArray,
    width: float = 0.5,
    dim: str | list[str] | None = None,
    greater_than: bool = True,
) -> xr.DataArray:
    """Calculate an extreme event probability.

    Args:
        x: Input DataArray of classifier values
        target: Target threshold to define an extreme.
        width: Transition width for a sigmoid function.
            If 0, use binary classification.
        dim: Optional dimension to reduce over. If provided, will calculate
            the mean probability across that dimension.
        greater_than: If `True`, classify as x >= target. If False, classify as x <= target.

    Returns:
        DataArray with values between 0 and 1 indicating probability of extreme event
    """
    attrs = x.attrs.copy()
    attrs["greater_than"] = greater_than
    if isinstance(target, float):
        attrs["target"] = target

    if width == 0:
        # xarray comparisons result in False where "x" is nan.  Preserve nans:
        result = xr.where(np.isnan(x), np.nan, (x >= target) if greater_than else (x <= target))
        result.name = "extreme"
        attrs.update({"long_name": "Event", "units": "T/F"})
    else:
        sign = 1 if greater_than else -1
        result = 1 / (1 + np.exp(-sign * (x - target) / width))
        result.name = "extreme_pct"
        attrs.update({"long_name": "Event probability", "units": "%", "width": width})

    if dim is not None:
        result = result.mean(dim=dim)

    result.attrs = attrs
    return result


def calibrate_event(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    greater_than: bool = True,
    groupby: str | list[str] | None = None,
) -> xr.DataArray:
    """Calculate calibrated extreme event probabilities using a calibrated logistic regression.

    Args:
        observed: Boolean DataArray indicating observed extreme events
        forecast: Continuous DataArray with forecast values
        groupby: Optional dimension(s) to group by when fitting logistic models.
            If None, fit a single model to all data.
        greater_than: If `True`, calibrate from upper quantiles.

    Returns:
        DataArray with calibrated probabilities between 0 and 1 indicating
        extreme event likelihood
    """
    result = (
        xr.zeros_like(observed, dtype=float)
        .rename("extreme_pct")
        .assign_attrs(forecast.attrs)
        .assign_attrs({"long_name": "Event Probability", "units": "%"})
    )
    qnt = [0.90, 0.75] if greater_than else [0.10, 0.25]

    if groupby is None:
        dim = [d for d in forecast.dims if d not in observed.dims]
        x = [forecast.quantile(q, dim=dim) for q in qnt] if dim else [forecast]

        # Sklearn expects non-NaN numpy (not xarray)
        x = np.column_stack([xcol.values.ravel() for xcol in x])
        y = observed.values.ravel()
        ok = ~np.isnan(y) & ~np.any(np.isnan(x), axis=1)
        # astype(int) fails on nan values, so we convert to int AFTER filtering:
        x_ok, y_ok = x[ok], y[ok].astype(int)

        y_sum = np.sum(y_ok)
        if y_sum == 0:  # Classifier fails if all observations are F
            probs = np.zeros_like(y)
        elif y_sum == len(y_ok):  # Classifier fails if all observations are T
            probs = np.ones_like(y)
        else:
            mdl = LogisticRegression(fit_intercept=True, solver="lbfgs")
            mdl.fit(x_ok, y_ok)
            probs = mdl.predict_proba(x)[:, 1]

        result.values = probs.reshape(observed.shape)
    else:
        # Calibrate each group individually and reassemble:
        groupby = [groupby] if isinstance(groupby, str) else groupby
        for group_key, group_fcst in forecast.groupby(groupby):
            sel_dict = dict(
                zip(groupby, group_key if isinstance(group_key, tuple) else [group_key])
            )
            group_obs = observed.sel(sel_dict)
            group_result = calibrate_event(group_obs, group_fcst, groupby=None)
            result.loc[sel_dict] = group_result

    return result


def build_confusion_matrix(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    reference: xr.DataArray | None = None,
    groupby: str | list[str] | None = None,
) -> xr.Dataset:
    """Calculate confusion matrix categories.

    Args:
        observed: Boolean `DataArray` indicating observed events
        forecast: Boolean `DataArray` indicating forecasted events
        reference: Optional boolean `DataArray` for paired confusion matrix
        groupby: Dimension(s) to group by/preserve (if `None`, calculate single score)

    Returns:
        Dataset containing observation percentages (0-1)
        If reference is None, simple confusion matrix:
            -`pp`: true positive (positive truth, positive forecast)
            -`nn`: true negative (negative truth, negative forecast)
            -`np`: false positive (negative truth, positive forecast)
            -`pn`: false negative (positive truth, negative forecast)
        If reference is provided, paired confusion matrix:
            -`nnn`: both true negative (negative truth, both negative)
            -`nnp`: reference false positive (negative truth, forecast negative, reference positive)
            -`npn`: forecast false positive (negative truth, forecast positive, reference negative)
            -`npp`: both false positive (negative truth, both positive)
            -`pnn`: both miss (positive truth, both negative)
            -`pnp`: reference true positive, forecast miss (positive truth, forecast negative, reference positive)
            -`ppn`: forecast true positive, reference miss (positive truth, forecast positive, reference negative)
            -`ppp`: both true positive (positive truth, both positive)
    """
    observed = observed.astype(bool)
    forecast = forecast.astype(bool)

    # Check if groupby is in coordinates but not dimensions, and swap if needed
    if groupby is not None:
        groupby_list = [groupby] if isinstance(groupby, str) else groupby
        for g in groupby_list:
            if g in observed.coords and g not in observed.dims:
                # Find the parent dimension of this coordinate
                parent_dim = [d for d in observed.dims if g in observed[d].coords][0]
                observed = observed.swap_dims({parent_dim: g})
                forecast = forecast.swap_dims({parent_dim: g})

    # Get model names from attributes, with fallbacks
    obs_name = getattr(observed, "model_name", "Observed")
    fcst_name = getattr(forecast, "model_name", "Forecast")

    if groupby is None:
        dims = observed.dims
    else:
        groupby = [groupby] if isinstance(groupby, str) else groupby
        dims = [dim for dim in observed.dims if dim not in groupby]

    total = observed.count(dim=dims)

    # Create filtered attributes without conflicting keys
    attrs = forecast.attrs.copy()
    attrs.pop("long_name", None)
    attrs.pop("units", None)
    attrs.pop("type", None)

    if reference is None:
        return xr.Dataset(
            {
                "nn": _safe_divide(
                    (~observed & ~forecast).sum(dim=dims),
                    total,
                    long_name="True Negative",
                    units="%",
                    type="tn",
                    **attrs,
                ),
                "np": _safe_divide(
                    (~observed & forecast).sum(dim=dims),
                    total,
                    long_name="False Positive",
                    units="%",
                    type="fp",
                    **attrs,
                ),
                "pn": _safe_divide(
                    (observed & ~forecast).sum(dim=dims),
                    total,
                    long_name="False Negative",
                    units="%",
                    type="fn",
                    **attrs,
                ),
                "pp": _safe_divide(
                    (observed & forecast).sum(dim=dims),
                    total,
                    long_name="True Positive",
                    units="%",
                    type="tp",
                    **attrs,
                ),
            }
        ).assign_attrs(
            observed_model_name=obs_name,
            forecast_model_name=fcst_name,
        )
    else:
        reference = reference.astype(bool)
        ref_name = getattr(reference, "model_name", "Reference")

        return xr.Dataset(
            {
                # Hit - True Negative
                "nnn": _safe_divide(
                    (~observed & ~forecast & ~reference).sum(dim=dims),
                    total,
                    long_name="Both True Negative",
                    units="%",
                    type="tn",
                    **attrs,
                ),
                # Three types of false positives:
                "nnp": _safe_divide(
                    (~observed & ~forecast & reference).sum(dim=dims),
                    total,
                    long_name=f"{ref_name} False Positive, {fcst_name} True Negative",
                    units="%",
                    type="fptn",  # false positive / true negative combo
                    **attrs,
                ),
                "npn": _safe_divide(
                    (~observed & forecast & ~reference).sum(dim=dims),
                    total,
                    long_name=f"{fcst_name} False Positive, {ref_name} True Negative",
                    units="%",
                    type="fptn",  # false positive / true negative combo
                    **attrs,
                ),
                "npp": _safe_divide(
                    (~observed & forecast & reference).sum(dim=dims),
                    total,
                    long_name="Both False Positive",
                    units="%",
                    type="fp",
                    **attrs,
                ),
                # Three types off False negatives (failure to detect):
                "pnn": _safe_divide(
                    (observed & ~forecast & ~reference).sum(dim=dims),
                    total,
                    long_name="Both False Negative",
                    units="%",
                    type="fn",
                    **attrs,
                ),
                "pnp": _safe_divide(
                    (observed & ~forecast & reference).sum(dim=dims),
                    total,
                    long_name=f"{ref_name} True Positive, {fcst_name} False Negative",
                    units="%",
                    type="fntp",  # false negative / true positive combo
                    **attrs,
                ),
                "ppn": _safe_divide(
                    (observed & forecast & ~reference).sum(dim=dims),
                    total,
                    long_name=f"{fcst_name} True Positive, {ref_name} False Negative",
                    units="%",
                    type="fntp",  # false negative / true positive combo
                    **attrs,
                ),
                # Hit - True Positive
                "ppp": _safe_divide(
                    (observed & forecast & reference).sum(dim=dims),
                    total,
                    long_name="Both True Positive",
                    units="%",
                    type="tp",
                    **attrs,
                ),
            }
        ).assign_attrs(
            observed_model_name=obs_name,
            forecast_model_name=fcst_name,
            reference_model_name=ref_name,
        )


def _calc_payoff(cm: xr.Dataset, payoff: dict | None = PAYOFF, **attrs) -> xr.Dataset | None:
    """Calculate payoff from a confusion matrix using provided coefficients.

    Args:
        cm: Confusion matrix Dataset containing data variables to score
        payoff: Dictionary mapping confusion matrix categories to their values
        **attrs: Additional attributes to add to the result

    Returns:
        Dataset containing:
            - Individual payoff components for each outcome
            - 'total': Sum of all outcome payoffs
        Returns None if `payoff` is None
    """
    if payoff is None:
        return None

    attrs["units"] = payoff.get("units", "$")
    coeffs = xr.Dataset({k: xr.DataArray(payoff.get(k, 0)) for k in cm.data_vars})
    with xr.set_options(keep_attrs=True):
        outcome = cm * coeffs
        for var in outcome:
            outcome[var].attrs["payoff"] = payoff.get(var, 0)
        attrs.update(long_name="Expected Payoff")
        attrs.update({k: payoff.get(k, 0) for k in cm.data_vars})
        outcome["total"] = outcome.to_array(dim="variable").sum("variable").assign_attrs(attrs)

    return outcome


def calc_f_score(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    reference: xr.DataArray | None = None,
    groupby: str | list[str] | None = None,
    beta: float = 1.0,
    payoff: dict = PAYOFF,
) -> xr.Dataset:
    """Calculate F-score components and optional payoff for event analysis.

    Args:
        observed: Boolean DataArray indicating observed extreme events
        forecast: Boolean DataArray indicating forecasted extreme events
        reference: Optional Boolean DataArray for calculating 4x2 paired payouts
        groupby: Dimension(s) to group by/preserve (if None, calculate single score)
        beta: Weight of recall in F-score calculation (default: 1.0 for F1 score)
        payoff: Optional dictionary with keys 'pp', 'nn', 'np', 'pn' specifying
            the value for each confusion matrix element. If provided, adds a
            'payoff' array to the output Dataset.

    Returns:
        Dataset containing:
            - `payoff`: Value calculated using payoff matrix
            - `f_score`: (1 + beta^2) * (precision * recall) / (beta^2 * precision + recall)
            - `precision`: True positives / (True positives + False positives, aka tpr)
            - `recall`: True positives / (True positives + False negatives)
            - `fpr`: False positive rate (False positives / Total negatives)
    """
    if reference is not None:
        cm_paired = build_confusion_matrix(observed, forecast, reference, groupby=groupby)
        po_paired = xr.Dataset({"payoff": _calc_payoff(cm_paired, payoff)["total"]})

        scores = [
            calc_f_score(observed, forecast, None, groupby, beta, payoff),
            calc_f_score(observed, reference, None, groupby, beta, payoff),
            po_paired,
        ]
        return xr.concat(
            scores, dim=pd.Index(["forecast", "reference", "paired"], name="forecast")
        ).assign_coords(
            model_name=(
                "forecast",
                [
                    getattr(forecast, "model_name", "Forecast"),
                    getattr(reference, "model_name", "Reference"),
                    "Paired",
                ],
            ),
            color=(
                "forecast",
                [
                    getattr(forecast, "color", "#1f77b4"),
                    getattr(reference, "color", "#ff7f0e"),
                    "#9467bd",  # "#2ca02c",
                ],
            ),
        )

    beta2 = beta**2
    cm = build_confusion_matrix(observed=observed, forecast=forecast, groupby=groupby)

    attrs = forecast.attrs.copy()
    attrs.pop("long_name", None)
    attrs.pop("units", None)

    payoff_value = _calc_payoff(cm, payoff, **attrs)
    if payoff_value is not None:
        payoff_value = payoff_value["total"]

    attrs["units"] = "%"
    precision = _safe_divide(cm.pp, cm.pp + cm.np, long_name="Precision", **attrs)
    recall = _safe_divide(cm.pp, cm.pp + cm.pn, long_name="Recall", alias="tpr", **attrs)
    fpr = _safe_divide(cm.np, cm.np + cm.nn, long_name="False Positive Rate", **attrs)
    f_score = _safe_divide(
        (1 + beta2) * (precision * recall),
        (beta2 * precision + recall),
        long_name=f"F{beta}-score",
        beta=beta,
        **attrs,
    )

    return xr.Dataset(
        {
            "payoff": payoff_value,
            "f_score": f_score,
            "precision": precision,
            "recall": recall,
            "fpr": fpr,
        }
    )


def style_confusion_matrix(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    reference: xr.DataArray | None = None,
    beta: float = 1,
    payoff: dict | None = None,
    ndigits: int = 2,
) -> Styler:
    """Create a styled DataFrame showing confusion matrix results.

    Args:
        observed: Boolean DataArray indicating observed events
        forecast: Boolean DataArray indicating forecasted events
        reference: Optional boolean DataArray for paired confusion matrix
        beta: Weight of recall in F-score calculation (default: 1.0 for F1 score)
        payoff: Payoff matrix.  If supplied (default none) also displays payoff values.
        ndigits: Number of digits of precision to display in the table

    Returns:
        Styled pandas DataFrame.
    """
    cm = build_confusion_matrix(observed, forecast, reference, groupby=None)
    fcst_name = getattr(forecast, "model_name", "Forecast")
    f_score = calc_f_score(observed, forecast, beta=beta).f_score.item()

    def mk_cell(type: str | list[str]):
        """Build the value to show in a cell of the table."""
        types = [type] if isinstance(type, str) else type
        pct = round(sum((100 * cm[t]).round(ndigits).values for t in types), ndigits)

        # Force grand total to exactly 100 to prevent near-100 rounding errors
        if set(types) == set(cm.data_vars):
            pct = 100.0

        if payoff is None:
            return f"<b>{pct}%</b>"
        else:
            value = round(sum(round(payoff[t] * cm[t].values, ndigits) for t in types), ndigits)
            pyoff = (
                payoff[types[0]]
                if len(types) == 1
                else round(value / (pct / 100) if pct != 0 else 0, ndigits)
            )
            return f"<b>{pct}%</b><br><small><i>*{pyoff} = {value}</i></small>"

    # Create data based on matrix type
    if reference is None:
        data = [
            [mk_cell("np"), mk_cell("pp"), mk_cell(["np", "pp"])],
            [mk_cell("nn"), mk_cell("pn"), mk_cell(["nn", "pn"])],
            [mk_cell(["np", "nn"]), mk_cell(["pp", "pn"]), mk_cell(list(cm.data_vars))],
        ]
        index = ["Extreme", "Normal", "Total"]
        cell_types = [
            [cm.np.attrs["type"], cm.pp.attrs["type"], "total"],
            [cm.nn.attrs["type"], cm.pn.attrs["type"], "total"],
            ["total", "total", "total"],
        ]
        caption = f"{fcst_name} F-score: {f_score:.3f}"
    else:
        ref_name = getattr(reference, "model_name", "Reference")
        ref_f_score = calc_f_score(observed, reference, beta=beta).f_score.item()

        data = [
            [mk_cell("npp"), mk_cell("ppp"), mk_cell(["npp", "ppp"])],
            [mk_cell("npn"), mk_cell("ppn"), mk_cell(["npn", "ppn"])],
            [mk_cell("nnp"), mk_cell("pnp"), mk_cell(["nnp", "pnp"])],
            [mk_cell("nnn"), mk_cell("pnn"), mk_cell(["nnn", "pnn"])],
            [
                mk_cell(["npp", "npn", "nnp", "nnn"]),
                mk_cell(["ppp", "ppn", "pnp", "pnn"]),
                mk_cell(list(cm.data_vars)),
            ],
        ]
        index = ["Both Extreme", f"{fcst_name} Only", f"{ref_name} Only", "Neither", "Total"]
        cell_types = [
            [cm.npp.attrs["type"], cm.ppp.attrs["type"], "total"],
            [cm.npn.attrs["type"], cm.ppn.attrs["type"], "total"],
            [cm.nnp.attrs["type"], cm.pnp.attrs["type"], "total"],
            [cm.nnn.attrs["type"], cm.pnn.attrs["type"], "total"],
            ["total", "total", "total"],
        ]
        caption = f"{fcst_name} F-score: {f_score:.3f}\n{ref_name} F-score: {ref_f_score:.3f}"

    df = pd.DataFrame(data, columns=["Normal", "Extreme", "Total"], index=index)
    df.columns.name = "Observed"
    df.index.name = "Forecast"

    color_map = {
        "fp": "background-color: rgba(255, 0, 0, 0.2)",  # red
        "fn": "background-color: rgba(255, 255, 0, 0.2)",  # yellow
        "tp": "background-color: rgba(0, 255, 0, 0.2)",  # green
        "tn": "background-color: rgba(0, 255, 0, 0.2)",  # green
        "total": "background-color: rgba(128, 128, 128, 0.2)",  # gray
        "fptn": "background: repeating-linear-gradient(45deg, rgba(255, 0, 0, 0.2) 0px, rgba(255, 0, 0, 0.2) 5px, rgba(0, 255, 0, 0.2) 5px, rgba(0, 255, 0, 0.2) 10px)",
        "fntp": "background: repeating-linear-gradient(45deg, rgba(255, 255, 0, 0.2) 0px, rgba(255, 255, 0, 0.2) 5px, rgba(0, 255, 0, 0.2) 5px, rgba(0, 255, 0, 0.2) 10px)",
    }

    styler = df.style.set_properties(**{"white-space": "pre-wrap"})

    def style_cells(x, cell_type: str):
        """Apply styling based on cell type."""
        mask = [[cell == cell_type for cell in row] for row in cell_types]
        return np.where(mask, color_map[cell_type], "")

    for cell_type in color_map:
        styler = styler.apply(style_cells, cell_type=cell_type, axis=None)

    return styler.set_caption(caption)


def search_threshold(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    objective: Literal["f_score", "payoff", "precision", "recall"] = "f_score",
    payoff: dict = PAYOFF,
    beta: float = 1,
    n_grid: int = 201,
) -> xr.Dataset:
    """Search for optimal thresholds by evaluating metrics across threshold values.

    Args:
        observed: Binary array of observed events (0 or 1)
        forecast: Numeric array of forecast probabilities
        objective: Metric to optimize ("f_score", "payoff", "precision", "recall")
        payoff: Dictionary specifying the value for each confusion matrix element
        beta: Weight parameter for F-score calculation
        n_grid: Number of threshold values to test

    Returns:
        Dataset containing all metrics for each threshold and the optimal threshold
    """
    # Create threshold grid adapted to forecast range
    TRIALS = "thresholds"
    thresholds = np.linspace(0, float(forecast.max().item()), n_grid)

    # Cartesian expansion of threshold and forecast's native
    threshold_dim = xr.DataArray(thresholds, dims=TRIALS, coords={TRIALS: thresholds})
    extreme = forecast >= threshold_dim

    # Calculate metrics across thresholds
    scores = calc_f_score(observed, extreme, groupby=TRIALS, payoff=payoff, beta=beta)
    scores = scores.assign_coords({TRIALS: thresholds})

    # Find optimal threshold
    objective_vals = scores[objective]
    optimal_idx = objective_vals.argmax(TRIALS).item()
    optimal_threshold = thresholds[optimal_idx]
    scores["threshold"] = xr.DataArray(optimal_threshold)
    scores["index"] = xr.DataArray(optimal_idx)

    # Handle metadata
    scores.attrs.update(forecast.attrs)
    scores.attrs.update({"objective": objective})
    scores.attrs.pop("long_name", None)
    scores.attrs.pop("units", None)

    return scores


def optimize_threshold(
    observed: xr.DataArray,
    forecast: xr.DataArray,
    objective: Literal["f_score", "payoff", "precision", "recall"] = "f_score",
    groupby: str | list[str] | None = None,
    payoff: dict = PAYOFF,
    beta: float = 1,
    n_grid: int = 201,
) -> xr.DataArray:
    """Find optimal thresholds for forecast probabilities that maximize an objective.

    Args:
        observed: Binary array of observed events (0 or 1)
        forecast: Numeric array of forecast probabilities
        objective: Metric to optimize ("f_score", "payoff", "precision", "recall")
        groupby: Dimension(s) to group by when computing thresholds
        payoff: Dictionary specifying the value for each confusion matrix element
        beta: Weight parameter for F-score calculation
        n_grid: Number of threshold values to test

    Returns:
        DataArray containing optimal thresholds for each group
    """
    if groupby is None:
        # If no groupby, just get the optimal threshold from search_threshold
        result = search_threshold(
            observed, forecast, objective=objective, payoff=payoff, beta=beta, n_grid=n_grid
        )["threshold"]
    else:
        # Convert groupby to list if it's a string
        groupby = [groupby] if isinstance(groupby, str) else groupby

        # Create output DataArray with dimensions from groupby
        coords = {dim: forecast[dim] for dim in groupby}
        result = xr.DataArray(np.nan, coords=coords, dims=groupby, name="threshold")

        # Process each group
        for group_key, group_data in forecast.groupby(groupby):
            sel_dict = dict(
                zip(groupby, group_key if isinstance(group_key, tuple) else [group_key])
            )

            # Get optimal threshold for this group using search_threshold
            result.loc[sel_dict] = search_threshold(
                observed.sel(sel_dict),
                group_data,
                objective=objective,
                beta=beta,
                payoff=payoff,
                n_grid=n_grid,
            )["threshold"].item()

    # Add metadata
    result.attrs.update(forecast.attrs)
    result.attrs.update(
        {
            "long_name": f"Optimal {forecast.attrs.get('long_name', '')} threshold",
            "objective": objective,
        }
    )

    if objective == "f_score":
        result.attrs["beta"] = beta
    elif objective == "payoff":
        result.attrs.update(
            {
                "payoff_pp": payoff.get("pp", 0),
                "payoff_nn": payoff.get("nn", 0),
                "payoff_np": payoff.get("np", 0),
                "payoff_pn": payoff.get("pn", 0),
            }
        )

    return result



#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Command line interface for the Salient SDK."""

import argparse
import os
from importlib import resources
from pathlib import Path

import pandas as pd
import xarray as xr

from .__init__ import __version__
from .data_timeseries_api import data_timeseries
from .downscale_api import downscale
from .forecast_timeseries_api import forecast_timeseries
from .geo_api import geo
from .hindcast_summary_api import hindcast_summary
from .location import Location
from .login_api import get_verify_ssl, login


def main() -> None:
    """Command line interface for the Salient SDK."""
    parser = _get_parser()

    # Convert arguments into action -------
    args = parser.parse_args()
    args_dict = vars(args)
    cmd = args_dict.pop("command")

    args_dict = _login_from_arg(args_dict)
    args_dict = _location_from_arg(args_dict)

    # Dispatch to the appropriate function based on the command
    if cmd == "forecast_timeseries":
        file_name = forecast_timeseries(**args_dict)
    elif cmd == "downscale":
        file_name = downscale(**args_dict)
    elif cmd == "data_timeseries":
        file_name = data_timeseries(**args_dict)
    elif cmd == "geo":
        file_name = geo(**args_dict)
    elif cmd == "hindcast_summary":
        file_name = hindcast_summary(**args_dict)
    elif cmd == "version":
        file_name = __version__
        args.verbose = False
    elif cmd == "login":
        file_name = args_dict["session"]
    elif cmd == "examples":
        file_name = "\n".join(_list_examples())
    else:
        # print(f"Command '{cmd}' not recognized")
        parser.print_help()
        file_name = None

    if file_name is None:
        pass
    elif (
        "verbose" in args
        and args.verbose
        and isinstance(file_name, str)
        and os.path.exists(file_name)
    ):
        ext = os.path.splitext(file_name)[1]
        if ext == ".nc":
            print(xr.open_dataset(file_name, decode_timedelta=True))
        elif ext == ".csv":
            print(pd.read_csv(file_name).head())
        else:
            print(file_name)
    else:
        print(file_name)


def _location_from_arg(arg: dict) -> dict:
    if all(key in arg for key in ["latitude", "longitude", "location_file", "shapefile"]):
        if not "region" in arg:
            # "region" is only used by hindcast_summary
            arg["region"] = None

        arg["loc"] = Location(
            lat=arg.pop("latitude"),
            lon=arg.pop("longitude"),
            location_file=arg.pop("location_file"),
            shapefile=arg.pop("shapefile"),
            region=arg.pop("region"),
        )

    return arg


def _login_from_arg(arg: dict) -> dict:
    if all(key in arg for key in ["username", "password", "apikey"]):
        arg["session"] = login(
            username=arg.pop("username"),
            password=arg.pop("password"),
            apikey=arg["apikey"],  # don't pop, we need this later
            verify=arg.pop("verify"),  # will reset momentarily
            verbose=arg["verbose"],  # don't pop, used elsewhere
        )
        # login() may set this as a side effect:
        arg["verify"] = get_verify_ssl()
    return arg


def _get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="salientsdk command line interface")
    subparsers = parser.add_subparsers(dest="command")

    # version command ---------
    subparsers.add_parser("version", help="Print the Salient SDK version")

    # examples command ---------
    subparsers.add_parser("examples", help="List example notebooks")

    # login command ----------
    _add_common_args(
        subparsers.add_parser("login", help="Login to the Salient API"),
        [],
        location_args=False,
        login_args=True,
        force=False,
    )

    # forecast_timeseries command ------------
    forecast_parser = _add_common_args(
        subparsers.add_parser("forecast_timeseries", help="Run the forecast_timeseries function"),
        ["date", "debias", "version", "variable"],
    )
    forecast_parser.add_argument("-fld", "--field", type=str, default="anom")
    forecast_parser.add_argument("-fmt", "--format", type=str, default="nc")
    forecast_parser.add_argument("-mdl", "--model", type=str, default="blend")
    forecast_parser.add_argument("-ref", "--reference_clim", type=str, default="30_yr")
    forecast_parser.add_argument("--timescale", type=str, default="all")

    # downscale command -----------
    downscale_parser = _add_common_args(
        subparsers.add_parser("downscale", help="Run the downscale function"),
        ["date", "debias", "version"],
    )
    # Note plural "variables", not "variable"
    downscale_parser.add_argument("-var", "--variables", type=str, default="temp,precip")
    downscale_parser.add_argument("--members", type=int, default=50)
    # These are currently the only available values.  Not necessary to expose
    # these as options from the command line.
    # argparser.add_argument("--format", type=str, default="nc")
    # argparser.add_argument("--frequency", type=str, default="daily")

    # data_timeseries command --------------
    data_parser = _add_common_args(
        subparsers.add_parser(
            "data_timeseries", help="Get historical ERA5 from the data_timeseries API"
        ),
        ["debias", "variable"],
    )
    data_parser.add_argument("-fld", "--field", type=str, default="anom")
    data_parser.add_argument("--start", type=str, default="1950-01-01")
    data_parser.add_argument("--end", type=str, default="-today")
    data_parser.add_argument("--format", type=str, default="nc")
    data_parser.add_argument("--frequency", type=str, default="daily")

    # hindcast_summary command ------------
    hnd = _add_common_args(
        subparsers.add_parser(
            "hindcast_summary",
            help="Get summary statistics for historical weather forecast quality",
        ),
        ["variable", "version", "force", "region"],
    )
    hnd.add_argument(
        "-met", "--metric", type=str, default="crps", help="The accuracy metric to calculate"
    )
    hnd.add_argument(
        "-ref",
        "--reference",
        type=str,
        default="-auto",
        help="The reference dataset to compare against",
    )
    hnd.add_argument("--season", type=str, default="all", help="The season to consider")
    hnd.add_argument("--split_set", type=str, default="all", help="The split set to use")
    hnd.add_argument("--timescale", type=str, default="all", help="Forecast look-ahead timescale")

    # geo command --------------
    geo_parser = _add_common_args(
        subparsers.add_parser("geo", help="Get static geo data from the geo API.")
    )
    geo_parser.add_argument(
        "-var",
        "--variables",
        type=str,
        default="temp,precip",
        help="Comma-separated list of variables to return.",
    )
    geo_parser.add_argument(
        "-fmt", "--format", type=str, default="nc", help="Format of the returned data."
    )
    geo_parser.add_argument(
        "-res",
        "--resolution",
        type=float,
        default=0.25,
        help="The spatial resolution of returned data in degrees.",
    )

    return parser


def _add_common_args(
    argparser: argparse.ArgumentParser,
    args: list[str] = [],
    location_args: bool = True,
    login_args: bool = True,
    force: bool = True,
) -> argparse.ArgumentParser:
    """Add standard arguments to a subparser.

    Args:
        argparser (argparse.ArgumentParser): The subparser to add arguments to.
        args (list[str]): Additional standard/shared arguments to add to the parser.
        location_args (bool): If True (default), add standard location arguments.
        login_args (bool): If True (default), add username/password arguments
        force (bool): If True (default), add a `--force` argument to overwrite existing files.
            Most API functions have this argument.

    Generate an argument parser for the Location class with consistent arguments
    that can be used across multiple `main()` functions and the command line.
    """
    if login_args:
        # Users will login with either a username/password or an API key:
        argparser.add_argument(
            "-u",
            "--username",
            type=str,
            default="SALIENT_USERNAME",  # pull from env var
            help="Salient-issued user name",
        )
        argparser.add_argument(
            "-p",
            "--password",
            type=str,
            default="SALIENT_PASSWORD",  # pull from env var
            help="Salient-issued password",
        )
        argparser.add_argument(
            "--apikey",
            type=str,  # Use "SALIENT_APIKEY" to pull from env var
            default=None,
            help="Salient-issued API key (ignores username/password)",
        )
        verify_group = argparser.add_mutually_exclusive_group(required=False)
        verify_group.add_argument(
            "--verify",
            dest="verify",
            action="store_true",
            help="Force verification of SSL certificates.",
        )
        verify_group.add_argument(
            "--noverify",
            dest="verify",
            action="store_false",
            help="Disable verification of SSL certificates.",
        )
        argparser.set_defaults(verify=None)

        verbosity_group = argparser.add_mutually_exclusive_group(required=False)
        verbosity_group.add_argument(
            "--verbose",
            dest="verbose",
            action="store_true",
            help="Print status messages (default behavior)",
        )
        verbosity_group.add_argument(
            "--quiet", dest="verbose", action="store_false", help="Suppress status messages"
        )
        argparser.set_defaults(verbose=True)

    if location_args:
        argparser.add_argument(
            "-lat",
            "--latitude",
            type=float,
            default=None,
            help="Decimal latitude -90 to 90 (also requires longitude)",
        )
        argparser.add_argument(
            "-lon",
            "--longitude",
            type=float,
            default=None,
            help="Decimal longitude -180 to 180 (also requires latitude)",
        )
        argparser.add_argument("-loc", "--location_file", type=str, default=None)
        argparser.add_argument("-shp", "--shapefile", type=str, default=None)

    if force:
        argparser.add_argument(
            "--force", action="store_true", help="Overwrite existing files (default is to cache)"
        )

    if "debias" in args:
        argparser.add_argument(
            "--debias",
            action="store_true",
            help="Debias to observation stations (default is no debiasing)",
        )

    if "version" in args:
        # Hidden argument, sunsetting "blend" model version selection
        argparser.add_argument(
            "-ver",
            "--model-version",
            default=None,
            dest="version",
            help=argparse.SUPPRESS,
        )

    if "date" in args:
        argparser.add_argument("--date", type=str, default="-today")

    if "region" in args:
        argparser.add_argument("--region", type=str, default=None)

    if "variable" in args:
        argparser.add_argument("-var", "--variable", type=str, default="temp")

    return argparser


def _list_examples():
    try:
        examples_path = resources.files("salientsdk") / "examples"
        return [
            str(Path(f).resolve())
            for f in examples_path.iterdir()
            if f.name.endswith(".ipynb") and not f.name.startswith(".")
        ]
    except Exception as e:
        raise FileNotFoundError(f"Could not find examples directory. Error: {e}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Whitelist file for vulture.

We use the vulture tool to make sure that we don't have dead code
hanging around. Sometimes vulture falsely flags a function as unused.
In that case, we add the function to this whitelist file so they are
explicitly registered as used.
"""

# from .upload_file_api import user_files
# assert user_files  # unused function (salientsdk/upload_file_api.py:220)

from .constants import set_model_version
from .forecast_zarr import ForecastZarr
from .hydro import exec_vic
from .solar import _downscale_solar

assert _downscale_solar  # currently under development and skipped in testing
assert exec_vic  # Ignore b/c VIC Image Driver is provided by user, not the SDK.
assert ForecastZarr.__new__  # Vulture doesn't like the factory class method
assert set_model_version  # Sunset feature, available via kwargs but not in public API



#!/usr/bin/env python
# Copyright Salient Predictions 2025

"""Interface to the Salient data_timeseries API.

Command line usage example:
```
cd ~/salientsdk
python -m salientsdk downscale -lat 42 -lon -73 --date 2020-01-01 -u username -p password --force
```

"""

import os
from datetime import datetime

import numpy as np
import pandas as pd
import requests
import xarray as xr

from .constants import _build_urls, _collapse_comma, _expand_comma
from .location import Location
from .login_api import download_queries

REFERENCE_CLIMS = ["30_yr", "10_yr", "5_yr", "salient"]
FREQUENCY = ["daily", "hourly"]

_EXCLUDE_ARGS = ["force", "session", "verify", "verbose", "destination", "loc", "strict", "kwargs"]


def downscale(
    # API arguments -----
    loc: Location,
    variables: str | list[str] = "temp,precip",
    date: str = "-today",
    members: int = 50,
    debias: bool = False,
    frequency="daily",
    reference_clim: str = "salient",
    weights: str | None = None,  # noqa
    length: int = 366,
    # Non-API arguments --------
    destination: str = "-default",
    force: bool = False,
    session: requests.Session | None = None,
    apikey: str | None = None,
    verify: bool | None = None,
    verbose: bool = False,
    strict: bool = True,
    **kwargs,
) -> str | pd.DataFrame:
    """Temporal downscale of forecasts.

    Convert temporally coarse probabilistic forecasts into granular daily ensembles.
    For more detail, see the
    [api doc](https://api.salientpredictions.com/v2/documentation/api/#/Forecasts/downscale).

    Args:
        loc (Location): The location to query.
            If using a `shapefile` or `location_file`, may input a vector of file names which
            will trigger multiple calls to `downscale`.  This is useful because `downscale` requires
            that all points in a file be from the same continent.
        variables (str | list[str]): The variables to query, separated by commas or as a `list`
            See the
            [Data Fields](https://salientpredictions.notion.site/Variables-d88463032846402e80c9c0972412fe60)
            documentation for a full list of available variables.
            Note that `downscale` natively supports a list of variables, so passing a list of
            variables here will not necessarily trigger downloading multiple files.
        date (str): The start date of the time series.
            If `date` is `-today`, use the current date.
        members (int): The number of ensemble members to download
        frequency (str): The temporal resolution of the time series, `daily` (default) or `hourly`.
        reference_clim (str): Reference period to calculate anomalies
        debias (bool): If True, debias the data to observation stations
        weights (str): The variable that will be used to weight each grid point within a
            `shapefile`.  Defaults to unweighted, can also weight by `population`.
        length (int): The number of days to downscale for. Defaults to 366.
        destination (str): The destination directory for downloaded files.
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        strict (bool): If True (default) error if query is invalid.  If False,
            return `NA` for the file name and continue processing.
        **kwargs: Additional arguments to pass to the API

    Keyword Arguments:
        gdd_base (int): The base temperature for growing degree days
        units (str): US or SI

    Returns:
        str | pd.DataFrame : If only one file was downloaded, return the name of the file.
            If multiple files were downloaded, return a table with column `file_name` and
            additional columns documenting the vectorized input arguments such as
            `location_file`.
    """
    if os.getenv("USE_MOCK_DOWNSCALE", "False") == "True":
        # pytest --nbmake does not support mock fixtures, requiring a more manual
        # approach to mocking when running notebooks as tests:
        try:
            return _downscale_mock(
                loc=loc,
                variables=variables,
                date=date,
                members=members,
                frequency=frequency,
                destination=destination,
                verbose=verbose,
                **kwargs,
            )
        except NotImplementedError as err:
            # mocking not available for all cases
            pass

    assert members > 0, "members must be a positive integer"
    assert reference_clim in REFERENCE_CLIMS, f"reference_clim must be one of {REFERENCE_CLIMS}"

    if not isinstance(length, int) or length <= 0:
        raise TypeError("length must be a positive integer")
    elif length > 366:
        raise ValueError("length must be less than 366 days")

    format = "nc"
    model = "blend"
    date = datetime.today().strftime("%Y-%m-%d") if date == "-today" else date
    variables = _collapse_comma(variables)
    frequency = _expand_comma(frequency, valid=FREQUENCY, name="frequency", default="daily")
    args = {k: v for k, v in {**locals(), **kwargs}.items() if k not in _EXCLUDE_ARGS}

    queries = _build_urls(endpoint="downscale", args=loc.asdict(**args), destination=destination)

    # return file_name because strict=False will NA failures
    queries["file_name"] = download_queries(
        query=queries["query"].values,
        file_name=queries["file_name"].values,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
        max_workers=5,  # downscale @limiter.limit("5 per second")
        strict=strict,
    )

    if len(queries) == 1:
        return queries["file_name"].values[0]
    else:
        # Now that we've executed the queries, we don't need it anymore:
        queries = queries.drop(columns="query")
        return queries


def _downscale_mock(
    loc: Location,
    variables: str | list[str] = "temp,precip",
    date: str = "2021-01-01",
    members: int = 21,
    # debias: bool = False,
    frequency="daily",
    # reference_clim: str = "salient",
    # version: str | list[str] = "-default",
    # weights: str | None = None,  # noqa
    # Non-API arguments --------
    destination: str = "-default",
    force: bool = False,
    # session: requests.Session | None = None,
    # apikey: str | None = None,
    # verify: bool | None = None,
    verbose: bool = False,
    **kwargs,
) -> str | pd.DataFrame:
    """Generate synthetic data quickly instead of calling the API."""
    format = "nc"
    model = "blend"
    date = datetime.today().strftime("%Y-%m-%d") if date == "-today" else date
    variables = _collapse_comma(variables)
    mock = True  # generate a different file_name than the real API call
    args = {k: v for k, v in {**locals(), **kwargs}.items() if k not in _EXCLUDE_ARGS}

    queries = _build_urls(endpoint="downscale", args=loc.asdict(**args), destination=destination)
    if len(queries) > 1:
        raise NotImplementedError("mock downscale expects scalar values")
    file_name = queries["file_name"].values[0]
    if os.path.exists(file_name) and not force:
        return file_name

    freq, time_cnt, tdim = (
        ("h", 8760, "time") if frequency == "hourly" else ("d", 365, "forecast_day")
    )
    variables = _expand_comma(variables)
    time = pd.date_range(start=date, periods=time_cnt, freq=freq)

    if loc.shapefile is not None:
        raise NotImplementedError("mock shapefile not available")
        # Shapefile downscales are shaped like:
        # (tdim,"ensemble","lat","lon")
    else:
        if loc.location_file is not None:
            geo = loc.load_location_file(destination=destination)
        elif loc.lat is not None:
            geo = pd.DataFrame({"name": ["0"], "lat": [loc.lat], "lon": [loc.lon]})
        loc_cnt = len(geo)

        data_vars = {
            var: ((tdim, "ensemble", "location"), np.random.rand(time_cnt, members, loc_cnt))
            for var in variables
        }
        coords = {
            tdim: (tdim, time),
            "ensemble": ("ensemble", np.array(range(members))),
            "location": ("location", geo["name"]),
            "lat": ("location", geo["lat"]),
            "lon": ("location", geo["lon"]),
            "forecast_date": np.datetime64(date, "ns"),
        }

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    if verbose:
        print(ds)
    ds.to_netcdf(file_name, encoding={"location": {"dtype": str}})
    return file_name
